/* som.c    main for som-language
	P. Chongstitvatana
	22 Oct 2002

	start Som project   			13 Sept 2003
	som v1  public release 			4 Dec 2004
	som-v15 rev2  seperate optim 	5 Mar 2005
	som-v16: hex, i-array, file i/o 6 Nov 2005
	som-v17 t-code					23 Dec 2005
	som-v17a						27 May 2006

*/

#include "compile.h"

extern  int line;		// current position in source

FILE *FO, *FI;			// output/input file
PRIVATE int def,im;		// list of function, immediate
PRIVATE int iMode;		// 1 interactive, 2 batch, 3 execute
int cmode;				// compile mode, 0 = def, 1 = im

extern int DS, GV;

void seterror( char *mess ){
	printf("\nError line %d near '",line);
	prToken(tok);
	printf("' : %s\n",mess);
//  dumpSymTab();
	exit(-1);
}

void warning( char *s, char *mess) {
#ifdef VERBOSE
	printf("Warning line %d, %s %s\n",line,s,mess);
#endif
}

PRIVATE char *stripdot(char *source){
	static char name[80];
	char *name2;
	strcpy(name,source);
	name2 = strtok(name,".");
	if( name2 == NULL ) name2 = name;
	return name2;
}

// parse, generate
PRIVATE void evalPrint(void){
	int e, tk;
	startlex();
	lex();
	cmode = 0;		// compile mode
	while( tok != tkEOF ){
		tk = tok;
		top();
		e = ypop();
		if(tk == tkTO || tk == tkCOLON){
//			def = append(def,e);  // for debugging
			genex(e);
		}else
			im = append(im,e);	// eval at runimm
	}
}

void loadfile( char *fn ){
	#ifdef VERBOSE
		printf("load %s\n",fn);
	#endif
	readinfile(fn);
	setcdr(im,NIL);		// clear im
	evalPrint();
	runimm(im);
}

// read from stdin, interactive mode
// read-eval-print loop
void rep(void){
	loadfile("lib.som");
	while(1){
		reader();
		setcdr(im,NIL);		// clear im
		evalPrint();
		runimm(im);
	}
}

void listing( char *source) {
	char *name;
	name = stripdot(source);
	strcat(name,".lst");
	FO = fopen(name,"wt");
	showCode(FO);
	fclose(FO);
}

void objfile( char *source) {
	char *obj;
	int i, k;
/*
	// done at dumpSymTab()
	for(i=0; i<nxsym; i++){		// update gvar
		k = exsym[i];
		if(getType(k) == tyGVAR)
			setLv(k,M[getRef(k)]);
	}
*/
	obj = stripdot(source);
	strcat(obj,".obj");
	FO = fopen(obj,"wt");
	fprintf(FO,"%d\n",OBJ_SOM17);
	outM(FO,CS,1,CP-1);			// code segment
	outM(FO,M,65,GV-1);			// globals
	outM(FO,M,MAXSYS,DS-1);		// data segment
	dumpSymTab(FO);
	fclose(FO);
}

void prolog(void){
	initCharType();
	initcell();
	initfile();
//	def = list(newatom(OPER,tkBB));
	im = list(newatom(OPER,tkBB));
}

// iMode 2 compile and run
void batch(char *source){
	loadfile("lib.som");
	loadfile(source);
//	prlist(def);
	genfinal();
//	showCode(stdout);
	nxsym = getExportSym(exsym);
	listing(source);
	objfile(source);
}

void execute(char *obj){
	int i, a;
	FI = fopen(obj,"r");
	fscanf(FI,"%d",&a);
	if( a != OBJ_SOM17 )
		seterror("not som 1.7 object");
	a = readM(FI,CS);	// code segment
	a = readM(FI,M);	// globals
	GV = a+1;
	a = readM(FI,M);	// data segment
	DS = a+1;			// free DS
/*
 test reLoc
 don't need to readSymTab if not reloc
	readSymTab(FI);
	nxsym = getExportSym(exsym);
	dumpSymTab(stdout);
	reLoc(1000);
*/
	fclose(FI);
	eval(1);
}

void Usage(void){
	printf("%s\n%s\n%s\n%s\n",
		"som          interactive mode",
		"som file     compile and run, output .lst and .obj",
		"som -x file  execute obj file",
		"som -?       this help");
	exit(0);
}

//extern int noi;

// iMode 1 interactive, 2 compile&run, 3 execute
int main( int argc, char *argv[] ){
	char source[32];
	int i;

	iMode = 2;
	for (i=1; i<argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case '?': Usage(); break;
			case 'x': iMode = 3; break;
			default:
				fprintf(stderr,"unknown option %s\n",argv[i]);
				exit(0);
			}
		}else
			break;   // first non option
	}
	if(argc == 1) iMode = 1;
	if(iMode == 2 || iMode == 3){
		// i is first non option arg
		strcpy(source,argv[i]);
	}
	prolog();
	switch(iMode){
	case 1:	rep();	break;
	case 2:	batch(source); break;
	case 3:	execute(source); break;
	}
//	printf("noi %d\n",noi);
	return 0;
}
